/****************************************************************************
 Module
   GameControlSM.c

 Revision
   2.0.1

 Description
   This is a template for the top level Hierarchical state machine

 Notes

****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for this state machine as well as any machines at the
   next lower level in the hierarchy that are sub-machines to this machine
*/
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "dbprintf.h"
#include "LeaderSM.h"
#include "CommSM.h"
#include "TimerSM.h"
#include "MotorControlSM.h"
#include "PIC32_SPI_HAL.h"
#include "PIC32PortHAL.h"

#include <stdio.h>
#include <stdint.h>
#include <sys/attribs.h>

/*----------------------------- Module Defines ----------------------------*/
#define HALF_SEC 500

#define SDIR SPI_RPA1
#define SDOR SPI_RPA2

#define MOTOR_CS 0b0000         // RA0
#define MOTOR_LAT _Pin_0
#define MOTOR_PORT _Port_A

#define LAUNCHER_CS 0b0011      // RB15
#define LAUNCHER_LAT _Pin_15
#define LAUNCHER_PORT _Port_B

#define MAP_CS1 0b0011

/*---------------------------- Module Functions ---------------------------*/
static void SendIt(uint8_t);
static void __ISR(_SPI_1_VECTOR, IPL7SOFT) ReceivedMsg(void);

/*---------------------------- Module Variables ---------------------------*/

static CommState_t CurrentState;

typedef enum {
    MotorPIC,
    LauncherPIC
}PICid_t;

static PICid_t PICtoCommand;

typedef enum {
    ReloadReached,
    LaunchReached,
    LaunchComplete
} Msg_t;

static uint8_t CurrentFollowerStatus;

static volatile uint32_t * const CS_Pins[2] = {&RPA0R, &RPB15R};
static uint32_t CS_IntMap[2] = {MOTOR_CS, LAUNCHER_CS};
static volatile uint32_t * const LAT_Reg[2] = {&LATA, &LATB};
static PortSetup_Pin_t LAT_Pin[2] = {MOTOR_LAT, LAUNCHER_LAT};
static PortSetup_Port_t Ports[2] = {MOTOR_PORT, LAUNCHER_PORT};

static PICid_t FollowerPICid;

/*------------------------------ Module Code ------------------------------*/
/****************************************************************************
Function
    RunCommSM

Parameters
   unsigned char: the event to process

Returns
   unsigned char: an event to return

Description
   add your description here
Notes
   uses nested switch/case to implement the machine.
Author
   Wing-Sum Law
****************************************************************************/
ES_Event_t RunCommSM(ES_Event_t CurrentEvent)
{
    bool MakeTransition = false;                    // Assume not making transition
	GameState_t NextState = CurrentState;           // Assume not making transition
	ES_Event_t ReturnEvent = CurrentEvent;          // Assume not consuming event
	ES_Event_t EntryEvent = {ES_ENTRY, 0};          // Entry event

    switch(CurrentState)
    {
        // If we are current waiting to query
        case WaitingToQuery:
        {
            switch(CurrentEvent.EventType)
            {
                // If event is ES_TIMEOUT: time to query relevant PIC
                case ES_TIMEOUT:
                {   
                    // Send query via SPI to query command generator
                    SendIt(Query);

                    // Switch state to Receiving
                    MakeTransition = true;
                    NextState = Receiving;
                }
                break;

                // If the beacon has been found
                case BeaconFound:
                {
                    #ifdef TERM_TESTING
                    DB_printf("Alignment complete, sending msg to move to launch\r\n");
                    #endif

                    // If beginning sweeping
                    if(CurrentEvent.EventParam == 0 || CurrentEvent.EventParam == 1)
                    {
                        // Send to motor PIC to move to launching position
                        SendIt(MoveToLaunch);

                    } else if(CurrentEvent.EventParam == 2)
                    {
                        // Send to motor PIC to stop moving
                        SendIt(StopTuning); 
                    }
                    
                }
                break;

                // If the game is over
                case GameOver:
                {
                    // Let active PIC know to stop
                    SendIt(StopEverything);
                    ResetTimerServo();
                }
            }
        }
        break;

        // If we are currently receiving a message
        case Receiving:
        {
            switch(CurrentEvent.EventType)
            {
                case MsgReceived:
                {
                    // Switch back to waiting to query
                    MakeTransition = true;
                    NextState = WaitingToQuery;
                    
                    // Restart query timer
                    ES_Timer_InitTimer(SPI_TIMER, HALF_SEC);
                }
                break;

                // If the tape has been found
                case BeaconFound:
                {   
                    #ifdef TERM_TESTING
                    DB_printf("Alignment complete, sending msg to move to launch\r\n");
                    #endif

                    // If beginning sweeping
                    if(CurrentEvent.EventParam == 0 || CurrentEvent.EventParam == 1)
                    {
                        // Send to motor PIC to move to launching position
                        SendIt(MoveToLaunch);

                    } else if(CurrentEvent.EventParam == 2)
                    {
                        // Send to motor PIC to stop moving
                        SendIt(StopTuning); 
                    }
                }
                break;

                // If the game is over
                case GameOver:
                {
                    // Let active PIC know to stop
                    SendIt(StopEverything);
                    ResetTimerServo();
                }
            }
        }
        break;
    }

    // If we are making a transition
    if(MakeTransition == true)
    {
        // Make the CurrentEvent an exit event and exit the current state
        CurrentEvent.EventType = ES_EXIT;
        RunCommSM(CurrentEvent);

        // Change the state variable to reflect the next state
        CurrentState = NextState;

        // Execute the entry function for the new state
        RunCommSM(EntryEvent);
    }

    // Always return same event unless error    
    return ReturnEvent;
}

/****************************************************************************
Function
     StartCommSM

Parameters
     ES_Event CurrentEvent

Returns
     nothing

Description
     Does any required initialization for this state machine, including SPI initialization
Notes

Author
     Wing-Sum Law
****************************************************************************/
void StartCommSM(ES_Event_t CurrentEvent)
{
    // Enter into the waiting to query state
    CurrentState = WaitingToQuery;

    // SPI SETUP HAPPENS HERE

    // Determine which PIC is to be follower
    uint8_t ThisCommand;
    if(CurrentEvent.EventParam == 0)
    {
        // Go to launch
        FollowerPICid = LauncherPIC;
        ThisCommand = StartLaunch;

    } else {
        // Go to motor
        FollowerPICid = MotorPIC;
        ThisCommand = (uint8_t)CurrentEvent.EventParam;
    }

    // Raise both SS lines just in case
    *LAT_Reg[0] |= LAT_Pin[0];
    *LAT_Reg[1] |= LAT_Pin[1];

    // Disable SPI interrupts while setting up
	IEC1CLR = _IEC1_SPI1RXIE_MASK;

	SPISetup_BasicConfig(SPI_SPI1);                            // Configure SPI1
    
    SPISetup_SetLeader(SPI_SPI1, SPI_SMP_MID);                 // Set Leader mode
    SPISetup_SetBitTime(SPI_SPI1, 10);                         // Set BRG using 
    
    SPISetup_SetActiveEdge(SPI_SPI1, SPI_SECOND_EDGE);         // Set CKE = 0 and CKP = 1
    SPISetup_SetClockIdleState(SPI_SPI1, SPI_CLK_HI);
    
    SPISetEnhancedBuffer(SPI_SPI1, false);                     // Disable enhanced buffer mode
    
    SPI1BUF = 0;                                               // Clear buffer
    
    // Map outputs
    SPISetup_MapSDInput(SPI_SPI1, SDIR);                       // Map SDI
	SPISetup_MapSDOutput(SPI_SPI1, SDOR);                      // Map SDO
    SPI1CONbits.MSSEN = 0;                                     // Disable auto driving of CS

    // Configuer CS lines
    PortSetup_ConfigureDigitalOutputs(Ports[FollowerPICid], LAT_Pin[FollowerPICid]);
    *CS_Pins[FollowerPICid] = MAP_CS1;                         // Map CS
    
    SPISetup_SetXferWidth(SPI_SPI1, SPI_8BIT);                 // Setup 8-bit transfers
    
	// Enable interrupts for receive and set priority to 7
	IFS1CLR = _IFS1_SPI1RXIF_MASK;							   // Clear SPI interrupt flag
	SPI1CONbits.SRXISEL = 1; 						  	       // When buffer not empty
	IPC7bits.SPI1IP = 7;
	IEC1SET = _IEC1_SPI1RXIE_MASK;							   // Reenable interrupts
    
    SPISetup_EnableSPI(SPI_SPI1);                              // Enable SPI1 again
    
    // Send command present to correct PIC
    SendIt(ThisCommand);
    
    #ifdef TERM_TESTING
    DB_printf("SPI Setup complete\r\n");
    #endif

    // Init SM
    RunCommSM(CurrentEvent);
    
    // Start query timer
    ES_Timer_InitTimer(SPI_TIMER, HALF_SEC);

    return;
}

/****************************************************************************
Function
     ReceivedMSG(void)

Parameters
     void

Returns
     void

Description
     Upon receiving message, post relevant event
Notes

Author
     Wing-Sum Law
****************************************************************************/
static void __ISR(_SPI_1_VECTOR, IPL7SOFT) ReceivedMsg(void)
{
    // Read SP1BUF to CurrentFollowerStatus
    CurrentFollowerStatus = (uint8_t)SPI1BUF;

    // Clear SPI1RXIF
    IFS1CLR = _IFS1_SPI1RXIF_MASK;
    
    #ifdef TERM_TESTING
    DB_printf("Received message: %d\r\n", CurrentFollowerStatus);
    #endif

    // Pass relevant event back to LeaderSM setting event param to msg
    // Send yourself message that msg was received
    ES_Event_t ThisEvent;
    ThisEvent.EventType = MsgReceived;
    ThisEvent.EventParam = CurrentFollowerStatus;

    // Raise CS line
    // Lower appropriate CS line
    *LAT_Reg[FollowerPICid] |= LAT_Pin[FollowerPICid];
    
    
    PostLeaderSM(ThisEvent);
}

static void SendIt(uint8_t msg)
{
    // Lower appropriate CS line
    *LAT_Reg[FollowerPICid] &= ~LAT_Pin[FollowerPICid];

    // Write msg to SPI buf
    SPI1BUF = msg;
}