/****************************************************************************
 Module
   GameControlSM.c

 Revision
   2.0.1

 Description
   This is a template for the top level Hierarchical state machine

 Notes

****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for this state machine as well as any machines at the
   next lower level in the hierarchy that are sub-machines to this machine
*/
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "dbprintf.h"
#include "GameControlSM.h"
#include "MotorControlSM.h"
#include "CommSM.h"
#include "PIC32PortHAL.h"

#include <stdio.h>

/*----------------------------- Module Defines ----------------------------*/

#define TEAMA_LED_PIN _Pin_10
#define LED_PORT _Port_B
#define TEAMB_LED_PIN _Pin_9

/*---------------------------- Module Functions ---------------------------*/

static ES_Event_t DuringMovingState(ES_Event_t Event);
static ES_Event_t DuringLaunchingState(ES_Event_t Event);
static ES_Event_t DuringReloadingState(ES_Event_t Event);

/*---------------------------- Module Variables ---------------------------*/

static GameOnState_t CurrentState;

/*------------------------------ Module Code ------------------------------*/
/****************************************************************************
Function
    RunGameControlSM

Parameters
   unsigned char: the event to process

Returns
   unsigned char: an event to return

Description
   add your description here
Notes
   uses nested switch/case to implement the machine.
Author
   Wing-Sum Law
****************************************************************************/
ES_Event_t RunGameControlSM(ES_Event_t CurrentEvent)
{
    bool MakeTransition = false;                    // Assume not making transition
	GameOnState_t NextState = CurrentState;         // Assume not making transition
	ES_Event_t ReturnEvent = CurrentEvent;          // Assume not consuming event
	ES_Event_t EntryEvent = {ES_ENTRY, 0};          // Entry event

    switch(CurrentState)
    {
        // if we are currently reloading
        case Reloading:
        {
            ReturnEvent = CurrentEvent = DuringReloadingState(CurrentEvent);
            // if we receive a reloading finished event
            if(CurrentEvent.EventType == DoneReloading)
            {
                // transition to moving
                NextState = Moving;
                MakeTransition = true;

                // Make sure transition to motor PIC control has correct entry
                EntryEvent.EventParam = MoveToLaunch;

                #ifdef TERM_TESTING
                DB_printf("Reloading complete, moving to launch position\r\n");
                #endif
            }
        }
        break;

        // if we are currently moving
        case Moving:
        {
            ReturnEvent = CurrentEvent = DuringMovingState(CurrentEvent);
            // If the CurrentEvent from the during function is an event
            if(CurrentEvent.EventType != ES_NO_EVENT)
            {
                switch(CurrentEvent.EventType)
                {
                    // If we have reached the launching position
                    case LaunchPositionReached:
                    {
                        #ifdef TERM_TESTING
                        DB_printf("Launch position reached, start launching \r\n");
                        #endif

                        // Transition to launching
                        NextState = Launching;
                        MakeTransition = true;
                    }
                    break;

                    // If we have reached the reload position
                    case ReloadPositionReached:
                    {
                        #ifdef TERM_TESTING
                        DB_printf("Reload position reached, wait for button input\r\n");
                        #endif
                        
                        // Transition to reloading
                        NextState = Reloading;
                        MakeTransition = true;
                    }
                    break;
                }
            }
        }
        break;

        // if we are currently launching
        case Launching:
        {
            ReturnEvent = CurrentEvent = DuringLaunchingState(CurrentEvent);

            // if we receive a launching finished event
            if(CurrentEvent.EventType == DoneLaunching)
            {
                // Make sure transition to motor PIC control has correct entry
                EntryEvent.EventParam = MoveToReload;

                #ifdef TERM_TESTING
                DB_printf("Launching complete, moving to reload position\r\n");
                #endif

                // Transition to moving
                NextState = Moving;
                MakeTransition = true;
            }
        }
        break;
            
    }
        

    // If we are making a transition
    if(MakeTransition == true)
    {
        // Make the CurrentEvent an exit event and exit the current state
        CurrentEvent.EventType = ES_EXIT;
        RunGameControlSM(CurrentEvent);

        // Change the state variable to reflect the next state
        CurrentState = NextState;

        // Execute the entry function for the new state
        RunGameControlSM(EntryEvent);
    }

    // Always return same event unless error    
    return ReturnEvent;
}

/****************************************************************************
Function
     StartGameControlSM

Parameters
     ES_Event CurrentEvent

Returns
     nothing

Description
     Does any required initialization for this state machine
Notes

Author
     Wing-Sum Law
****************************************************************************/
void StartGameControlSM(ES_Event_t CurrentEvent)
{
    // Enter into the moving (aligning) state
    CurrentState = Moving;

    // Configure LED pins
    PortSetup_ConfigureDigitalOutputs(LED_PORT, TEAMA_LED_PIN | TEAMB_LED_PIN);

    // Make sure both are zero-ed
    LATB = LATB & ~TEAMA_LED_PIN; 
    LATB = LATB & ~TEAMB_LED_PIN;

    // Init SM
    RunGameControlSM(CurrentEvent);

    return;
}

/****************************************************************************
 Function
     QueryTemplateSM

 Parameters
     None

 Returns
     TemplateState_t The current state of the Template state machine

 Description
     returns the current state of the Template state machine
 Notes

 Author
     J. Edward Carryer, 10/23/11, 19:21
****************************************************************************/
GameOnState_t QueryGameControlSM(void)
{
    return CurrentState;
}

/***************************************************************************
private functions
***************************************************************************/
/****************************************************************************
Function
     DuringMovingState

Parameters
     ES_Event CurrentEvent

Returns
     ES_Event_t

Description
     During the moving stage, control MovingSM (and therefore Moving PIC)
Notes

Author
     Wing-Sum Law
****************************************************************************/
static ES_Event_t DuringMovingState(ES_Event_t Event)
{
    ES_Event_t ReturnEvent = Event;

    // If we are entering the state
    if(Event.EventType == ES_ENTRY)
    {
        // Start the SPI (which will query the Moving PIC) -- order important!
        // Set up the motor communication line by setting event param to first 8bits as 0 to indicate which ss
        ES_Event_t StartCommEvent = {ES_ENTRY, Event.EventParam};
        StartCommSM(StartCommEvent);

        // Start the MovingSM (which will control the Moving PIC) into correct state
        // Contains event param indicating type of entry
        StartMotorControlSM(Event);

    } else if(Event.EventType == ES_EXIT)
    {
        // Exit the MovingSM
        // Exit SPI
        RunCommSM(Event);
        RunMotorControlSM(Event);
    } else {
        // No during functions
        // Pass events on to the lower states
        RunCommSM(Event);
        RunMotorControlSM(Event);
    }
       
    return ReturnEvent;
}

/****************************************************************************
Function
     DuringLaunchingState

Parameters
     ES_Event CurrentEvent

Returns
     ES_Event_t

Description
     During the waiting stage, do nothing unless start button is newly pressed
Notes

Author
     Wing-Sum Law
****************************************************************************/
static ES_Event_t DuringLaunchingState(ES_Event_t Event)
{
    ES_Event_t ReturnEvent = Event;

    // If we are entering the launching state
    if(Event.EventType == ES_ENTRY)
    {
        // Start the SPI (which will query the Launching PIC) -- order important!
        ES_Event_t StartCommEvent = {ES_ENTRY, 0};
        StartCommSM(StartCommEvent);

        // Start the LaunchingSM (which will control the Launching PIC)
        
    } else if(Event.EventType == ES_EXIT)
    {
        // Exit SPI
        RunCommSM(Event);
    } else {
        
        RunCommSM(Event);

        if(Event.EventParam == 2) // If we are done launching
        {
            ReturnEvent.EventType = DoneLaunching;
        }
    }

    // No during function for this state

    return ReturnEvent;
}

/****************************************************************************
Function
     DuringReloadingState

Parameters
     ES_Event CurrentEvent

Returns
     ES_Event_t

Description
     During the waiting stage, do nothing unless start button is newly pressed
Notes

Author
     Wing-Sum Law
****************************************************************************/
static ES_Event_t DuringReloadingState(ES_Event_t Event)
{
    ES_Event_t ReturnEvent = Event;

    // If entering the Reloading state
    if(Event.EventType == ES_ENTRY) {
        // Just waiting
    } else if(Event.EventType == ES_EXIT)
    {
        // Will start moving again
    }

    // No during function for this state

    return ReturnEvent;
}