/****************************************************************************
 Module
   LeaderSM.c

 Revision
   2.0.1

 Description
   This is a template for the top level Hierarchical state machine

 Notes

****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for this state machine as well as any machines at the
   next lower level in the hierarchy that are sub-machines to this machine
*/
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "LeaderSM.h"
#include "TimerSM.h"
#include "GameControlSM.h"
#include "PIC32PortHAL.h"
#include "CommSM.h"
#include "dbprintf.h"

#include <stdio.h>

/*----------------------------- Module Defines ----------------------------*/

#define StartButtonPort _Port_B
#define StartButtonPin _Pin_3
#define ReloadButtonPort _Port_B
#define ReloadButtonPin _Pin_4

#define TEAMA_LED_PIN _Pin_10
#define LED_PORT _Port_B
#define TEAMB_LED_PIN _Pin_9

/*---------------------------- Module Functions ---------------------------*/

static ES_Event_t DuringWaitingState(ES_Event_t Event);
static ES_Event_t DuringGameOnState(ES_Event_t Event);

/*---------------------------- Module Variables ---------------------------*/

static GameState_t CurrentState;
static uint8_t MyPriority;

/*------------------------------ Module Code ------------------------------*/
/****************************************************************************
Function
     InitLeaderSM

Parameters
     uint8_t : the priorty of this service

Returns
     bool, false if error in initialization, true otherwise

Description
     Saves away the priority, and starts
     the top level state machine
Notes

Author
     Wing-Sum
****************************************************************************/
bool InitLeaderSM(uint8_t Priority)
{
    // Make a new ES_ENTRY event
    ES_Event_t ThisEvent;
    ThisEvent.EventType = ES_ENTRY;

    // Save Priority as MyPriority
    MyPriority = Priority;

    // Configure start button pin
    PortSetup_ConfigureDigitalInputs(StartButtonPort, StartButtonPin);

    // Configure reload button pin
    PortSetup_ConfigureDigitalInputs(ReloadButtonPort, ReloadButtonPin);

    // Start this state machine by passing ES_ENTRY event
    StartLeaderSM(ThisEvent);

    // Return true (assumes no errors if it makes it through StartLeaderSM)
    return true;
}

/****************************************************************************
 Function
     PostLeaderSM

 Parameters
     ES_Event_t ThisEvent , the event to post to the queue

 Returns
     boolean False if the post operation failed, True otherwise

 Description
     Posts an event to this state machine's queue
 Notes

 Author
     J. Edward Carryer, 10/23/11, 19:25
****************************************************************************/
bool PostLeaderSM( ES_Event_t ThisEvent )
{
  return ES_PostToService(MyPriority, ThisEvent);
}

/****************************************************************************
Function
    RunLeaderSM

Parameters
   unsigned char: the event to process

Returns
   unsigned char: an event to return

Description
   add your description here
Notes
   uses nested switch/case to implement the machine.
Author
   Wing-Sum Law
****************************************************************************/
ES_Event_t RunLeaderSM(ES_Event_t CurrentEvent)
{
    bool MakeTransition = false;                    // Assume not making transition
    GameState_t NextState = CurrentState;           // Assume not making transition
    ES_Event_t ReturnEvent = {ES_NO_EVENT, 0} ;     // Assume no event
    ES_Event_t EntryEvent = {ES_ENTRY, 0};          // Entry event

    switch(CurrentState)
    {
        // If we are currently waiting
        case Waiting:
        {
            // Execute during function (processes entry and exit)
            CurrentEvent = DuringWaitingState(CurrentEvent);

            // If the CurrentEvent from the during function is an event
            if(CurrentEvent.EventType != ES_NO_EVENT)
            {
                switch(CurrentEvent.EventType)
                {
                    // If the start button has been pressed, enter game mode
                    case StartButtonPressed:
                    {
                        #ifdef TERM_TESTING
                        DB_printf("Start button pressed\r\n");
                        #endif

                        NextState = GameOn;
                        MakeTransition = true;
                    }
                    break;
                }
            }
        }
        break;

        // If we are currently running the game
        case GameOn:
        {
            // Execute during function (processes entry and exit)
            CurrentEvent = DuringGameOnState(CurrentEvent);

            // If CurrentEvent is an event
            if(CurrentEvent.EventType != ES_NO_EVENT)
            {
                switch(CurrentEvent.EventType)
                {
                    // If time is out and it is game over, go back to waiting
                    case GameOver:
                    {
                        #ifdef TERM_TESTING
                        DB_printf("Game over!\r\n");
                        #endif

                        ResetTimerServo();

                        // Turn off LEDs
                        LATB = LATB & ~TEAMA_LED_PIN; 
                        LATB = LATB & ~TEAMB_LED_PIN;

                        NextState = Waiting;
                        MakeTransition = true;
                    }
                    break;
                }
            }
        }
        break;
    }

    // If we are making a transition
    if(MakeTransition == true)
    {
        #ifdef TERM_TESTING
        DB_printf("Making transition\r\n");
        #endif

        // Make the CurrentEvent an exit event and exit the current state
        CurrentEvent.EventType = ES_EXIT;
        RunLeaderSM(CurrentEvent);

        // Change the state variable to reflect the next state
        CurrentState = NextState;

        // Execute the entry function for the new state
        RunLeaderSM(EntryEvent);
    }

    // Always return no event unless error    
    return ReturnEvent;
}

/****************************************************************************
Function
     StartLeaderSM

Parameters
     ES_Event CurrentEvent

Returns
     nothing

Description
     Does any required initialization for this state machine
Notes

Author
     Wing-Sum Law
****************************************************************************/
void StartLeaderSM(ES_Event_t CurrentEvent)
{
    // Enter into the Waiting state
    CurrentState = Waiting;

    // Init SMs
    RunLeaderSM(CurrentEvent);
    return;
}

/***************************************************************************
private functions
***************************************************************************/
/****************************************************************************
Function
     DuringWaitingState

Parameters
     ES_Event CurrentEvent

Returns
     ES_Event_t

Description
     During the waiting stage, do nothing unless start button is newly pressed
Notes

Author
     Wing-Sum Law
****************************************************************************/
static ES_Event_t DuringWaitingState(ES_Event_t Event)
{
    ES_Event_t ReturnEvent = Event;

    // If we are entering the state
    if(Event.EventType == ES_ENTRY)
    {
        // Continue waiting

    } else if(Event.EventType == ES_EXIT)
    {
        // Deactivate start button checker
        // WaitingForStart = false;
    }

    // Else, continue waiting
    // No during function for this state
       
    return ReturnEvent;
}

/****************************************************************************
Function
     DuringGameOnState

Parameters
     ES_Event CurrentEvent

Returns
     ES_Event_t

Description
     During the GameOn stage, wait until time is up
Notes

Author
     Wing-Sum Law
****************************************************************************/
static ES_Event_t DuringGameOnState(ES_Event_t Event)
{
    ES_Event_t ReturnEvent = Event;

    // If entering the GameOn state
    if(Event.EventType == ES_ENTRY)
    {
        // Initialize lower GameControlSM
        // Include start sweeping param to get passed down to motor
        Event.EventParam = StartSweeping;
        
        #ifdef TERM_TESTING
        DB_printf("Passed start sweeping command to GameControlSM\r\n");
        #endif

        StartGameControlSM(Event);
    } else if(Event.EventType == ES_EXIT)
    {
        // Exit from the bottom up
        ReturnEvent = RunGameControlSM(Event);

    } else
    {
        // Continue running lower level machines
        // No during function for this state
        ReturnEvent = RunGameControlSM(Event);
    }

    return ReturnEvent;
}

/****************************************************************************
 Function
     QueryTemplateSM

 Parameters
     None

 Returns
     TemplateState_t The current state of the Template state machine

 Description
     returns the current state of the Template state machine
 Notes

 Author
     J. Edward Carryer, 10/23/11, 19:21
****************************************************************************/
GameState_t QueryLeaderSM(void)
{
  return CurrentState;
}
