/****************************************************************************
 Module
   MotorControlSM.c

 Revision
   2.0.1

 Description
   This is a template for the top level Hierarchical state machine

 Notes

****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for this state machine as well as any machines at the
   next lower level in the hierarchy that are sub-machines to this machine
*/
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "MotorControlSM.h"
#include "GameControlSM.h"
#include "CommSM.h"
#include "LeaderSM.h"
#include "PIC32_SPI_HAL.h"
#include "PIC32PortHAL.h"
#include "dbprintf.h"

#include <stdio.h>
#include <stdint.h>
#include <sys/attribs.h>

/*----------------------------- Module Defines ----------------------------*/

#define HALF_SEC 250

#define ICPinPORT _Port_A
#define ICPinNum _Pin_4

#define IC1_PINMAPPING 0b0010
#define PBCLK_FREQ 20000000

#define TEAMA_PR 750   // # ticks in 300 us
#define TEAMB_PR 2750  // # ticks 1100 us
#define TOLERANCE 45
#define TUNING_TOLERANCE 25

#define TEAMA_LED_PIN _Pin_10
#define LED_PORT _Port_B
#define TEAMB_LED_PIN _Pin_9

/*---------------------------- Module Functions ---------------------------*/
static void TimerSetup(void);
static void ICSetup(void);
static void __ISR(_INPUT_CAPTURE_1_VECTOR, IPL7SOFT) InputCaptureISR(void);
static void __ISR(_TIMER_2_VECTOR, IPL6SOFT) Timer2ISR(void);

/*---------------------------- Module Variables ---------------------------*/

static MotorControlState_t CurrentState;

static uint32_t CurrentTime;
static uint32_t LastTime;

static uint8_t MyTeam;
static uint32_t TeamPRs[2] = {TEAMA_PR, TEAMB_PR};

union Timer {
    struct ByWord {
        uint16_t CapturedTime;
        uint16_t RolloverCount;
    } MyWord;
    uint32_t CurrentTime;
}MyTimer;

/*------------------------------ Module Code ------------------------------*/
MotorControlState_t QueryMotorControlSM(void)
{
    return CurrentState;
}

/****************************************************************************
Function
    RunMotorControlSM

Parameters
   unsigned char: the event to process

Returns
   unsigned char: an event to return

Description
   add your description here
Notes
   uses nested switch/case to implement the machine.
Author
   Wing-Sum Law
****************************************************************************/
ES_Event_t RunMotorControlSM(ES_Event_t CurrentEvent)
{
    bool MakeTransition = false;                    // Assume not making transition
	GameOnState_t NextState = CurrentState;         // Assume not making transition
	ES_Event_t ReturnEvent = CurrentEvent;          // Assume not consuming event
	ES_Event_t EntryEvent = {ES_ENTRY, 0};          // Entry event

    switch(CurrentState)
    {
        // If we are currently sweeping & looking for beacon
        case Sweeping:
        {
            // If the beacon has been found
            if(CurrentEvent.EventType == BeaconFound)
            {
                #ifdef TERM_TESTING
                DB_printf("Beacon found, looking for tape\r\n");
                #endif

                // Determine beacon frequency and light up relevant team
                if(CurrentEvent.EventParam == 0) {
                    LATB = LATB | TEAMA_LED_PIN;
                    MyTeam = 0;
                    
                    #ifdef TERM_TESTING
                    DB_printf("Team A\r\n");
                    #endif

                } else if(CurrentEvent.EventParam == 1) {
                    LATB = LATB | TEAMB_LED_PIN;
                    MyTeam = 1;
                    
                    #ifdef TERM_TESTING
                    DB_printf("Team B\r\n");
                    #endif
                }
                
                // Move to moving to launch stage
                NextState = MovingToLaunch;
                MakeTransition = true;
            }
        }
        break;

        // If we are currently moving to launch position
        case MovingToLaunch:
        {
            // If moving to launch has completed
            if(CurrentEvent.EventType == MsgReceived && CurrentEvent.EventParam == 4)
            {
                // Alert leader SM that launch position has been reached
                ES_Event_t ThisMsg = {LaunchPositionReached, 0};
                PostLeaderSM(ThisMsg);

                #ifdef TERM_TESTING
                DB_printf("Launch position reached, leaving moving state from MotorControlSM\r\n");
                #endif
//                if(CurrentEvent.EventParam == 4)
//                {
//                    // Start looking for beacon again
//                    // Turn on IC1 and T2CON
//                    T2CONbits.ON = 1;
//                    IC1CONbits.ON = 1;
//
//                    // Move to tuning stage
//                    NextState = Tuning;
//                    MakeTransition = true;
//                }
            }
        }
        break;

        // If we are currently tuning shooting alignment
        case Tuning:
        {
            // If tuning is complete
            if(CurrentEvent.EventType == BeaconFound)
            {
                // Alert leader SM that launch position has been reached
                ES_Event_t ThisMsg = {LaunchPositionReached, 0};
                PostLeaderSM(ThisMsg);

                #ifdef TERM_TESTING
                DB_printf("Launch position reached, leaving moving state from MotorControlSM\r\n");
                #endif
            }
        }
        break;

        // If we are currently moving to reload position
        case MovingToReload:
        {
            // If moving to reload has completed
            if(CurrentEvent.EventType == MsgReceived)
            {
                if(CurrentEvent.EventParam == 1)
                {
                    // Alert leader SM that launch position has been reached
                    ES_Event_t ThisMsg = {ReloadPositionReached, 0};
                    PostLeaderSM(ThisMsg);

                    #ifdef TERM_TESTING
                    DB_printf("Reload position reached, leaving moving state from MotorControlSM\r\n");
                    #endif
                }
            }
        }
        break;

    }

    // If we are making a transition
    if(MakeTransition == true)
    {
        // Make the CurrentEvent an exit event and exit the current state
        CurrentEvent.EventType = ES_EXIT;
        RunMotorControlSM(CurrentEvent);

        // Change the state variable to reflect the next state
        CurrentState = NextState;

        // Execute the entry function for the new state
        RunMotorControlSM(EntryEvent);
    }

    // Always return same event unless error    
    return ReturnEvent;
}

/****************************************************************************
Function
     StartMotorControlSM

Parameters
     ES_Event CurrentEvent

Returns
     nothing

Description
     Determines which state we're entering, and therefore what to do
Notes

Author
     Wing-Sum Law
****************************************************************************/
void StartMotorControlSM(ES_Event_t CurrentEvent)
{   
    // Determine which state to enter into
    // if entering sweeping state
    if(CurrentEvent.EventParam == StartSweeping)
    {
        // Set up input capture
        TimerSetup();
        ICSetup();

        #ifdef TERM_TESTING
        DB_printf("IC Setup complete\r\n");
        #endif
        
        // Start timer & input capture
        TMR2 = 0;
        T2CONbits.ON = 1;
        IC1CONbits.ON = 1;

        // Entering sweeping state
        CurrentState = Sweeping;

    } else if(CurrentEvent.EventParam == MoveToLaunch)
    {
        // Enter moving to launch state
        CurrentState = MovingToLaunch;

    } else if(CurrentEvent.EventParam == MoveToReload)
    {
        // Enter moving to reload state
        CurrentState = MovingToReload;
    }

    // Init SM
    RunMotorControlSM(CurrentEvent);

    return;
}

/***************************************************************************
private functions
***************************************************************************/
/****************************************************************************
Function
     TimerSetup(void)

Parameters
     void

Returns
     void

Description
     Set up timers for input capture

Notes

Author
     Wing-Sum Law
****************************************************************************/
static void TimerSetup(void)
{
    //Disable Timer 2
    T2CONbits.ON = 0;

    //Select the internal PBCLK source
    T2CONbits.TCS = 0;

    //Clear timer gate
    T2CONbits.TGATE = 0;

    //Set prescale to 8
    T2CONbits.TCKPS = 0b011;

    //Set match count (period) to 0xFFFF
    PR2 = 0xFFFF;

    //Enable local interrupt (rollover)
    IEC0SET = _IEC0_T2IE_MASK;

    //Make sure we are set up for multiple interrupt vectors
    INTCONbits.MVEC = 1;

    //Configure interrupt priority to be 6
    IPC2bits.T2IP = 6;
}

static void ICSetup(void)
{
    /* INPUT CAPTURE SETUP */
    //Assign pin RPA4 to Timer 2
    PortSetup_ConfigureDigitalInputs(ICPinPORT, ICPinNum);

    IC1R = IC1_PINMAPPING; //0b0010

    //Select 16-bit timer resource capture
    IC1CONbits.C32 = 0;

    //Set timer 2 for input capture
    IC1CONbits.ICTMR = 1;

    //Interrupt on every capture event
    IC1CONbits.ICI = 0;

    //Set mode to detect only rising edges
    IC1CONbits.ICM = 0b011;

    //Enable input capture interrupts
    IEC0SET = _IEC0_IC1IE_MASK;

    //Set input capture interrupt priority to 7
    IPC1bits.IC1IP = 7;
}

static void __ISR(_INPUT_CAPTURE_1_VECTOR, IPL7SOFT) InputCaptureISR(void)
{    
    do{
        //Read IC1BUF into CapturedTime
        MyTimer.MyWord.CapturedTime = (uint16_t)IC1BUF;

        //Calculate period by subtracting LastTime from CurrentTime
        uint32_t PeriodTicks = MyTimer.CurrentTime - LastTime;
        
        #ifdef TERM_TESTING
        DB_printf("Entered IC PeriodTicks %d\r\n", PeriodTicks);
        #endif

        // If we are sweeping at the beginning
        if(CurrentState == Sweeping)
        {
            // If calculated frequency is near Team A frequency
            if ((PeriodTicks >= (TEAMA_PR - TOLERANCE)) && (PeriodTicks <= (TEAMA_PR + TOLERANCE)))
            {
                ES_Event_t ThisEvent;
                ThisEvent.EventType = BeaconFound;
                ThisEvent.EventParam = 0;

                // Turn off IC1 and T2CON
                T2CONbits.ON = 0;
                IC1CONbits.ON = 0;

                PostLeaderSM(ThisEvent);
                
            } else if ((PeriodTicks >= (TEAMB_PR - TOLERANCE)) && (PeriodTicks <= (TEAMB_PR + TOLERANCE)))
            {
                ES_Event_t ThisEvent;
                ThisEvent.EventType = BeaconFound;
                ThisEvent.EventParam = 1;

                // Turn off IC1 and T2CON
                T2CONbits.ON = 0;
                IC1CONbits.ON = 0;
                
                PostLeaderSM(ThisEvent);
            }

        } else if(CurrentState == Tuning)
        {
            if((PeriodTicks >= (TeamPRs[MyTeam] - TUNING_TOLERANCE)) && (PeriodTicks <= (TeamPRs[MyTeam] + TUNING_TOLERANCE)))
            {
                ES_Event_t ThisEvent;
                ThisEvent.EventType = BeaconFound;
                ThisEvent.EventParam = 2;

                PostLeaderSM(ThisEvent);

                // Turn off IC1 and T2CON
                T2CONbits.ON = 0;
                IC1CONbits.ON = 0;
            }
        }
        

        //Update LastTime to be CurrentTime
        LastTime = MyTimer.CurrentTime;
        

    }while(IC1CONbits.ICBNE != 0); //Until the input capture buffer is empty

    //Clear IC interrupt
    IFS0CLR = _IFS0_IC1IF_MASK;

    //If Timer 2 has an interrupt flag pending
    if (IFS0bits.T2IF == 1 && MyTimer.MyWord.CapturedTime < 0x800)
    {
        //Increment RollOverCounter
        ++MyTimer.MyWord.RolloverCount;

        //Clear the Rollover Interrupt
        IFS0CLR = _IFS0_T2IF_MASK;
    }
}

/* Rollover Timer 2 Interrupt ISR */

static void __ISR(_TIMER_2_VECTOR, IPL6SOFT) Timer2ISR(void)
{
    //Disable interrupts globally
    __builtin_disable_interrupts();

    //IF T2IF pending
    if (IFS0bits.T2IF == 1) {

        //Increment RolloverCount
        ++MyTimer.MyWord.RolloverCount;

        //Clear rollover interrupt
        IFS0CLR = _IFS0_T2IF_MASK;

    }

    //Re-enable interrupts
    __builtin_enable_interrupts();

}
