/****************************************************************************
 Module
   TemplateFSM.c

 Revision
   1.0.1

 Description
   Times the game and controls timer servo

 Notes

****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for this state machine as well as any machines at the
   next lower level in the hierarchy that are sub-machines to this machine
*/
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "TimerSM.h"
#include "LeaderSM.h"
#include "dbprintf.h"

/*----------------------------- Module Defines ----------------------------*/

#define ONE_SEC 1000      // assumes 1 ms / tick
#define TOTAL_TIME 138    // number of seconds on the clock

#define TICKS_PER_MS 2500 // 20 MHz PBCLK

// Extents of the TimerServo motion
#define FINISHED ((uint16_t)(1*TICKS_PER_MS)) //0.55
#define WAITING ((uint16_t)(2*TICKS_PER_MS))

// Full range over 2:18 or 138 s
#define TICKS_PER_STEP ((WAITING - FINISHED)/138)
#define SERVO_PERIOD (20*TICKS_PER_MS) // 50 Hz

/*---------------------------- Module Functions ---------------------------*/
/* prototypes for private functions for this machine.They should be functions
   relevant to the behavior of this state machine
*/

static void ServoSetup(void);
static void MoveTimerServo(void);


/*---------------------------- Module Variables ---------------------------*/
// everybody needs a state variable, you may need others as well.
// type of state variable should match htat of enum in header file
static TimerState_t CurrentState;

// with the introduction of Gen2, we need a module level Priority var as well
static uint8_t MyPriority;

// Timer for the game
static uint8_t TicksRemaining = TOTAL_TIME;

static uint16_t CurrentPosition;

/*------------------------------ Module Code ------------------------------*/
/****************************************************************************
Function
     InitTimerSM

Parameters
     uint8_t : the priorty of this service

Returns
     bool, false if error in initialization, true otherwise

Description
     Saves away the priority, and enters WaitingToTime
Notes

Author
     Wing-Sum
****************************************************************************/
bool InitTimerSM(uint8_t Priority)
{
    // Make a new ES_INIT event
    ES_Event_t ThisEvent;

    // Save Priority as MyPriority
    MyPriority = Priority;

    // Setup timer servo
    ServoSetup();
    ResetTimerServo();

    // Post event to service to transition into WaitingToTime state
    ThisEvent.EventType = ES_INIT;
    if (ES_PostToService(MyPriority, ThisEvent) == true)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/****************************************************************************
 Function
     PostTimerSM

 Parameters
     ES_Event_t ThisEvent , the event to post to the queue

 Returns
     boolean False if the post operation failed, True otherwise

 Description
     Posts an event to this state machine's queue
 Notes

 Author
     J. Edward Carryer, 10/23/11, 19:25
****************************************************************************/
bool PostTimerSM( ES_Event_t ThisEvent )
{
  return ES_PostToService(MyPriority, ThisEvent);
}

/****************************************************************************
Function
    RunTimerSM

Parameters
   unsigned char: the event to process

Returns
   unsigned char: an event to return

Description
   add your description here
Notes
   uses nested switch/case to implement the machine.
Author
   Wing-Sum Law
****************************************************************************/
ES_Event_t RunTimerSM(ES_Event_t CurrentEvent)
{
    ES_Event_t ReturnEvent;
    ReturnEvent.EventType = ES_NO_EVENT; // assume no event returned

    switch(CurrentState)
    {
        // If we are currently in the init pseudostate
        case InitTimerState:
        {
            if(CurrentEvent.EventType == ES_INIT)
            {
                // Transition to waiting to start timing
                CurrentState = WaitingToCount;
            }
        } 
        break;

        // If we are currently waiting before game
        case WaitingToCount:
        {
            // If the start button has been pressed
            if(CurrentEvent.EventType == StartButtonPressed)
            {
                // Start timer
                ES_Timer_InitTimer(GAME_TIMER, ONE_SEC);

                // Set servo
                // SetTimerServo();

                // Transition to counting state
                CurrentState = Counting;

            }
        }
        break;

        // If the game is currently running
        case Counting:
        {
            // Whenever a single sec has passed
            if(CurrentEvent.EventType == ES_TIMEOUT)
            {
                // Lower remaining time
                TicksRemaining -= 1;
                
                // If there is no time remaining in the game
                if(TicksRemaining == 0)
                {
                    // Reset TimerServo
                    ResetTimerServo();
                    
                    // Alert the LeaderSM that the game is over
                    ES_Event_t GameOverEvent;
                    GameOverEvent.EventType = GameOver;

                    PostLeaderSM(GameOverEvent);

                    // Return to a waiting state
                    CurrentState = WaitingToCount;
 
                    // Reset total timer
                    TicksRemaining = TOTAL_TIME;

                } else {
                    // Move Servo and Restart timer
                    ES_Timer_InitTimer(GAME_TIMER, ONE_SEC);

                    // Move servo
                    MoveTimerServo();
                }
            }

        }
        break;

    }

    return ReturnEvent;
}

// Private function for starting up OC for timer servo

static void ServoSetup(void) {

    // Turn off timer 2
    T3CONbits.ON = 0;

    // Base timer2 on PBCLK/8
    T3CONbits.TCS = 0;
    T3CONbits.TCKPS = 0b011;
    PR2 = SERVO_PERIOD;

    // Set up output control
    OC2CONbits.ON = 0;

    // Select timer 2
    OC2CONbits.OCTSEL = 1;

    // Map RB15 to OC2
    RPB5R = 0b0101;

    // Set PWM mode no fault detect
    OC2CONbits.OCM = 0b110;

    // Default to DC 0
    OC2R = 0;

    OC2CONbits.ON = 1;

    // Turn timer 2 on
    T3CONbits.ON = 1;
}

static void MoveTimerServo(void)
{
    CurrentPosition -= TICKS_PER_STEP;

    // Update DC
    OC2RS = CurrentPosition;

    // Set timer
    ES_Timer_InitTimer(GAME_TIMER, ONE_SEC);

}

void ResetTimerServo(void)
{
    CurrentPosition = WAITING;
    printf("RESET TIMER SERVO\n\r");

    OC2RS = WAITING;
}

void SetTimerServo(void)
{
    printf("SET TIMER SERVO\r\n");
    CurrentPosition = FINISHED;

    OC2RS = FINISHED;
}